#!/bin/sh
# ============================================================================
#
# This file is part of the 'FFP-stick for ZyXEL NSA-nnn'
# http://zyxel.nas-central.org/wiki/FFP-stick
# 
# Adapted to also support the Iomega lifeline base NASses, april 2013
#
# This program is free software; you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# Author: Mijzelf <Mijzelf@live.com>
#
# ============================================================================
#
# This script has multiple functions.
# - Download, unpack and start ffp
# - Stop ffp
# - Execute any function outside the ffp chroot
#
# It is supposed to be started by usb_key_func.sh/fun_plug.sh or by itself.
# In case of a zypkg it is started by /usr/local/zy-pkg/etc/init.d/ffp
# In case of an lifeline package it is started by /opt/apps/Mijzelf/ffp.sh
#
# Execute the script without params for a brief help

# ===========================================
# Options 
# ===========================================
# Choose which ffp version to use. 0.7/arm, 0.7/oarm, 0.5, 0.6 or auto. 
# Currently 'auto' means 0.7/arm for all boxes which support EABI, and 0.7/oarm for all other boxes
FFPVERSION=auto
#
# Choose if you want ffp chrooted. (no, yes, auto)
# Currently 'auto' means 'yes' for arm boxes (fw>=4.40) and 'no' for oabi boxes (you'll need the /ffp symlink too).
FFPCHROOT=auto
# (The script will chroot anyway if the root is read only, or if you 
# didn't enable the /ffp symlink
# if chrooted
    # Choose if you want /i-data and /e-data to be symlinked (yes or no)
    # When not symlinked the directory will be doubled
    FFPSYMLINKDATA=yes
    #
# else
    # Choose if you want the root and admin shell 
    # changed to /ffp/bin/sh if not chrooted.
    FFPCHANGEROOTSHELL=yes
    #
# fi

# Choose if you want sshd auto-enabled on installation
# Currently 'auto' means 'yes' for zypkg and lifeline and 'no' for stick
FFPENABLE_SSH=auto

# ===========================================
# End of options
# ===========================================
SCRIPTVERSION="20131124"

# To be filled by commandline options:
FFPROOT=
NASTYPE=
FIRMWARE=
CHROOTDIR=
FFPTYPE=stick
#

HELPERDIR=/tmp/.ffpstick/
HELPERFILE=helper.inc

# ===================================================
# function SplitPath
# split path in first parameter in path and filename
SPLITTEDPATH="" # Output path
SPLITTEDFILE="" # Output file
# environment: current directory changes
# 
# Could of course also be done with basename and 
# dirname, but I didn't know about that functions 
# when I first wrote this script. This works, and I
# don't know if those function are available on all
# targets
# ===================================================
SplitPath()
{
        SPLITTEDPATH=""
        SPLITTEDFILE=""
        IFSBACKUP=${IFS};   IFS=/                       
                   
        for chunk in $1; do
    		newpath="" 
                if [ "${SPLITTEDPATH}" = "" ] ; then
            		newpath="${chunk}"
                else        
            		newpath="${SPLITTEDPATH}/${chunk}"
                fi          
    		if [ "${newpath}" = "$1" ] ; then
            		SPLITTEDFILE="${chunk}"
                elif [ "/${newpath}" = "$1" ] ; then
            		SPLITTEDPATH="/${SPLITTEDPATH}"
            		SPLITTEDFILE="${chunk}"
                else        
            		SPLITTEDPATH="${newpath}"
        	fi          
        done                                            
	IFS=${IFSBACKUP}    
	if [ "" != "${SPLITTEDPATH}" ]; then
	    cd ${SPLITTEDPATH}
	fi                  
	SPLITTEDPATH=` pwd `
}                           

# =============================================================
# function CleanUp 
# Delete some unneeded (and possibly dangerous) DNS-323 scripts
# =============================================================
CleanUp()
{
    echo -ne Delete some DNS-323 specific stuff...
    for file in start/passwd.sh start/rcS.sh start/shells.sh sbin/store-passwd.sh start/fp-master.sh sbin/fp.master
    do
	rm -f ${FFPROOT}/ffp/${file}
    done
    
    # ========================================================
    # Added 20120924
    # /ffp/start/sshd.sh in FFP 0.7 contains a little bug
    # ========================================================
    sed -i 's|-f \/etc\/ssh|-r \/ffp\/etc\/ssh|g' ${FFPROOT}/ffp/start/sshd.sh

    echo done
}

# =============================================================
# function IsAbiSupported
# Checks if oabi or eabi is supported
# Uses the external executable testeabi or testoabi
# 1st argument: test or bg (internal use)
# 2nd argument eabi or oabi
# =============================================================
IsAbiSupported()
{
    if [ "$1" = "bg" ]; then
	echo no >/tmp/$2.status
	if ${FFPROOT}/bin/test$2 2>/dev/null ; then
	    echo yes >/tmp/$2.status
	fi
	return
    fi

    IsAbiSupported bg $2 &
    sleep 5
    killall test$2 >/dev/null 2>&1
    supported=` cat /tmp/$2.status `
    rm /tmp/$2.status
    if [ "${supported}" = "no" ] ; then
	return 1
    fi
    return 0
}

# =============================================================
# function Unpack_fun_plug
# Unpack fonz' base package
# 1st argument: filename
# =============================================================
Unpack_fun_plug()
{
    echo Unpack ${FFPROOT}/$1...
    cd ffp
    tar -xzf ../$1
    cd ..
    rm $1
    echo Done
    CleanUp
}

# =============================================================
# function CheckForFfpTarball
# Check for a user-provided ffp base package either in /path/to/ffproot
# or it's parent (the ext3 partition of the ffp-stick)
# =============================================================
CheckForFfpTarball()
{
    # find ffp tarball
    for file in fun_plug.tgz ../fun_plug.tgz
    do
	if [ -r ${file} ]; then
	    Unpack_fun_plug ${file}
	fi
    done
}

# =============================================================
# function TryToDownload
# Downloads ffp base package from inreto.de
# =============================================================
TryToDownload()
{
#############################################
# Option for ffp 0.6 added at 2012-01-10
# as version 0.5 denied to run on the
# Medion P89626/P89630
# Option for ffp 0.7 added at 2012-01-31
# Option for ffp 0.7/oabi added at 2012-03-06
# Option for ffp.embedded added at 2013-11-25
#############################################
    if [ -d ${FFPROOT}/ffp.embedded ] ; then
        # An embedded ffp was provided with the FFP package
        echo using embedded FFP
        if [ -d ${FFPROOT}/ffp ] ; then
            rm -rf ${FFPROOT}/ffp
        fi

	mv ${FFPROOT}/ffp.embedded ${FFPROOT}/ffp
	return 0
    fi
    #############################################
    FFP05=http://ffp.inreto.de/ffp/0.5/
    FFP06=http://ffp.inreto.de/testing/ffp/0.6/arm/
    FFP07arm=http://ffp.inreto.de/ffp/0.7/arm/
    FFP07oarm=http://ffp.inreto.de/ffp/0.7/oabi/

    # Check if a stock wget is available:
    WGET=` which wget `
    
    if [ "" = "${WGET}" ] ; then
	# if not, use the one from the stick
	WGET=${FFPROOT}/bin/wget
    fi

    if ! [ -x ${WGET} ]; then
	echo No wget found
	return 0
    fi

    FFPURL=${FFP05}
    
    case ${FFPVERSION} 
    in
	0.5*)	FFPURL=${FFP05}
		;;
	0.6*)	FFPURL=${FFP06}
		;;
	0.7/arm*)
		FFPURL=${FFP07arm}
		;;
	0.7/oarm*)
		FFPURL=${FFP07oarm}
		;;
	*)	echo Which FFP version to use?	
		echo -ne Is OABI supported?...
		if IsAbiSupported test oabi ; then
		    echo -ne "yes. "
		else
		    echo -ne "no. "
		fi
		echo -ne Is EABI supported?...
		if IsAbiSupported test eabi ; then
		    echo yes, choosing FFP 0.7/arm
		    FFPURL=${FFP07arm}
		    FFPVERSION=0.7/arm
		else
		    echo no, choosing FFP 0.7/oarm
		    FFPURL=${FFP07oarm}
		    FFPVERSION=0.7/oarm
		fi
		;;
    esac

    FFPURL=${FFPURL}fun_plug.tgz
    echo Will try to download fun_plug.tgz \(version ${FFPVERSION}\)
    echo ${FFPURL}
    echo If the script stops here, downloading the tarball from ffp.inreto.de failed.
    if [ "$FFPTYPE" != "zypkg" ] && [ "$FFPTYPE" != "lifeline" ] ; then
        echo You can try to get internet access for the ${NASTYPE},
        echo "or copy fun_plug.tgz (case sensitive!) to the share containing ffproot."
        echo It should be recognized and installed.
        echo It is possible that you have to reboot the box before this share shows up.
    fi
    echo ping ffp.inreto.de...
    while ! [ -r ffp/bin/busybox ] && ! ping -c 1 ffp.inreto.de >/dev/null ; do
	sleep 10
	CheckForFfpTarball
    done  

    if ! [ -r ffp/bin/busybox ]; then
	echo ffp.inreto.de found. Downloading fun_plug.tgz...
	cd ffp
	echo pwd ` pwd `

	${WGET} -qO- ${FFPURL} | tar -xz
	if [ "$?" != "0" ] ; then
    	    echo Download failed.
	    if [ "$FFPTYPE" = "zypkg" ] || [ "$FFPTYPE" = "lifeline" ] ; then
		exit 0
	    fi
	else
	    echo fun_plug.tgz is downloaded
	    CleanUp
	fi
    	cd ..
    fi
    
    SITESFILE=ffp/etc/funpkg/sites
    if [ -r ${SITESFILE} ] ; then
	FFPVERSION2=${FFPVERSION}
	if [ "${FFPVERSION2}" = "0.7/oarm" ]; then
	    FFPVERSION2=0.7/oabi
	fi
    
	echo "mz	http://downloads.zyxel.nas-central.org/Users/Mijzelf/FFP-Stick/packages/${FFPVERSION2}" >> ${SITESFILE}
    fi
    
    # Put the ffp version in the etc directory:
    echo FFPVERSION=${FFPVERSION} >ffp/etc/ffp.version
    
    # Say hello to the author of this script. This doesn't do anything, 
    # (the file doesn't exist, so you'll get a 404), except generating 
    # a logline on my webserver. (When it's online), Then I know that yet 
    # another FFP-stick is created. I like that. ;)
    # If you want to opt-out, just remove the line below
    ${WGET} -qO- http://mijzelf.hopto.org/hello_author.php?nas=${NASTYPE},${FIRMWARE}\&ffp=${FFPVERSION}-${FFPTYPE},${SCRIPTVERSION} >/dev/null 2>&1
    if [ -r ./bin/wget ] ; then
        rm ./bin/wget
    fi
}

#####################
# Function added 20101217 to solve the problem that the /i-data/md0
# symlink was not copied right on the NSA-320
#####################
CopySymlink()
{
#    echo CopySymlink $1 $2
    target=` ls -l $1 | awk -F\  '{print $NF}' `    
#    echo ln -s $target $2
    ln -s $target $2
}

# ===========================================================
# function UmountTree
# Traverse a tree, and umount all mountpoints
# ===========================================================
UmountTree()
{
#    echo UmountTree $1
    if [ -h $1 ] ; then
	# Directory turns out to be a symlink. 
	return 0
    fi
    if mountpoint $1 >/dev/null 2>&1
    then
	# Directory is a mountpoint
	echo -ne umount $1...
	if umount -l $1
	then
	    echo done
	    return 0
	fi
	
	echo failed!
	return 1
    fi

    for file in ` ls -A $1 `
    do
	if [ -d $1/${file} ] ; then
	# File is a directory
	    if ! UmountTree $1/${file} 
	    then
		return 1
	    fi
        fi
    done
    return 0
}

# ===========================================================
# function CopyTree
# Helper function for the NSA3x0.
# Copies a complete directory tree, using ln or cp
# symlinks and mountpoints are treated special
# ===========================================================
CopyTree() 
{
    if [ -h $2 ] ; then
	# Directory turns out to be a symlink. Just copy it
	echo " (symlink)"
	if [ "$1" != "nop" ]; then
	    CopySymlink $2 $3
	fi
	return 0
    fi
    
    mkdir -p $3
    if mountpoint $2 >/dev/null 2>&1
    then
	# Directory is a mountpoint
	echo " (mountpoint)"
	mount --bind $2 $3
	
	if [ "0" != "$?" ] ; then
	    return 1
	fi
	return 0
    fi 
    echo
    for file in ` ls -A $2 `
    do
	echo -ne $4${file}
	if [ -d $2/${file} ] ; then
	# File is a directory
	    echo -ne " (dir)"
	    CopyTree $1 $2/${file} $3/${file} "$4----"
	    if [ "0" != "$?" ] ; then
	        return 1
	    fi
	#########################################
	# Added 2011-03-08; Changed 2011-03-23;
	# The NSA-320 has a symlink /etc/exports, 
	# which points to a non-existing file
	# as long as the nfs packet is not installed.
	# This frustrated this copy script
	elif [ -h $2/${file} ] ; then
	    # File is a symlink. Just copy it
	    echo " (symlink)"
	    if [ "$1" != "nop" ]; then
		CopySymlink $2/${file} $3/${file}
	    fi
	#########################################
	else
	    case $1 in
	    	nop)
		    echo
		    ;;
		ln)
		    echo
		    ln $2/${file} $3/${file}
		    ;;
		cp)
		    echo
		    cp -p $2/${file} $3/${file}
		    ;;
	    esac
	    if [ "0" != "$?" ] ; then
		return 1
	    fi 
	fi
    done
    return 0
}

# ===================================================================
# function SymlinkDir
# Moves a directory, and leaves a symlink in place
# Arguments:
# $1 source directory
# $2 target directory
# ===================================================================
SymlinkDir()
{
    if [ "${FFPSYMLINKDATA:-yes}" = "no" ] ; then
	return 1
    fi

    echo -ne Trying to rename /$1...
    if ! mv /$1 /$1.old 
    then
	echo Failed. Read only rootfs?
	return 1
    fi
    echo done
    mkdir -p $2/$1
    for dir in ` ls -A /$1.old/  `
    do
	if [ -h /$1.old/$dir ] ; then
	    echo "    Copy symlink /$1/$dir"
	    CopySymlink /$1.old/$dir $2/$1/$dir
	    rm /$1.old/$dir
	else
	    mkdir -p $2/$1/$dir
	    echo -ne "    Move /$1/$dir..."
	    if mount --move /$1.old/$dir $2/$1/$dir
	    then
		rmdir /$1.old/$dir
		echo done
	    else
		echo failed
		mv /$1.old /$1
		return 1
	    fi
	fi
    done
    
    ln -s $2/$1 /$1
    rmdir /$1.old
    echo done
    return 0
}

# ===================================================================
# function NSA320BindAll
# Helper function for the NSA320. Because the firmware directories cannot
# be bind-mounted on the FFP-stick directories, we'll inverse it. 
# Basically the rootfs is doubled using hardlinks, and the 2nd copy is 
# chrooted, with the FFP-stick bindmounted on /2ndcopy/ffp
# ===================================================================
NSA320BindAll()
{
    local chdir=/tmp/bindroot
    local action=ln
    local lndirs="bin sbin lib usr dev mnt"

    if [ -d ${chdir} ] && [ -d ${chdir}/i-data ] ; then
	echo I think the FFP tree was already built
	lndirs="$lndirs e-data i-data"
	action=nop
    else
	mkdir -p ${chdir}

	if SymlinkDir e-data ${chdir}
	then
	    if ! SymlinkDir i-data ${chdir}
	    then
		lndirs="$lndirs i-data"
	    fi
	else
	    lndirs="$lndirs e-data i-data"
	fi
    fi

    echo Hardlink the root tree
    for dir in $lndirs
    do
	echo -ne CopyTree $action /${dir} ${chdir}/${dir}
	CopyTree $action /${dir} ${chdir}/${dir} "----"
	if [ "0" != "$?" ]; then
	    echo Failed to CopyTree /${dir}
	    return 1
	fi
    done

    # Maybe we are a zypkg? Then we need
    if [ "${FFPTYPE}" = "zypkg" ]; then
	mount --bind /usr/local/zy-pkgs ${chdir}/usr/local/zy-pkgs
    fi

    local restartsamba=no

    echo -ne "Checking if /etc is a symlink..."
    if [ -h /etc ] ; then
	echo "it is. We'll leave it alone."
    else
	
    
	echo "it isn't. We'll hijack it."
        mkdir -p ${chdir}/tmp
	mount -t tmpfs tmfs ${chdir}/tmp
	if [ "0" != "$?" ]; then
	    echo Failed to mount tmpfs on ${chdir}/tmp
    	    return 1
	fi


	echo Create a copy of /etc
	echo -ne CopyTree ln /etc ${chdir}/.etc
	CopyTree ln /etc ${chdir}/.etc "----"
	if [ "0" != "$?" ]; then
	    echo Failed to CopyTree /etc
	    return 1
	fi

	echo ` date ` Waiting for fadd to show up...
	while ! ps | grep /sbin/fadd >/dev/null ; do
	    sleep 10
	done
	echo ` date ` Hello fadd!

# Because the files in /etc are bound to be changed by the firmware, we can't just make hardlinks to
# it. We'll just hijack the map...
	echo ...and hijack the original /etc...
	mv /etc /.etc.old
	ln -s ${chdir}/.etc /etc
#    mount --bind ${chdir}/.etc /etc
        if [ "0" != "$?" ]; then
	    echo Failed to hijack /etc
    	    return 1
	fi
    
	echo /etc/ has been exchanged by a symlink to ${chdir}/.etc
	restartsamba=yes
    fi

    echo -ne "Check if /var/log is a symlink..."
    if [ -h /var/log ] ; then
	echo "It is. Leave it alone"
    else
	echo "It isn't. We'll hijack it"
	# We'll hijack /var/log to, for the same reason...
	mkdir -p ${chdir}/var/log
	echo -ne CopyTree ln /var/log ${chdir}/var/log
	CopyTree ln /var/log ${chdir}/var/log "----"
	mv /var/log /var/.log.old
	ln -s ${chdir}/var/log /var/log 
	echo done
	restartsamba=yes
    fi
    
    # Restart Samba because it has gone mad about the /etc exchange
    if [ "restartsamba" = "yes" ] ; then
	echo Restart Samba
	/etc/init.d/samba.sh restart
    fi

    if [ -d /.etc.old ] ; then
	echo umount all mounts beneath /etc
	if UmountTree /.etc.old
	then
	    echo ...and rm -rf /.etc.old
	    rm -rf /.etc.old
	fi
    fi

    if [ -d /var/.log.old ] ; then
	if UmountTree /var/.log.old
	then
	    rm -rf /var/.log.old
	fi
    fi

    echo "Final symlinks and mounts"
    ln -sf ${FFPROOT}/ffp ${chdir}/ffp
    ln -sf ${FFPROOT}/ffp/etc ${chdir}/etc
    ln -sf ${FFPROOT}/${SPLITTEDFILE} ${chdir}/${SPLITTEDFILE}

    echo Create convenience links /ffp/etc/original and /ffp/etc/zyxel
    for dir in original zyxel
    do
	local shortdir=${chdir}${FFPROOT}/ffp/etc/$dir
	if [ -d $shortdir ] ; then
	    rmdir $shortdir
	elif [ -f $shortdir ] ; then
	    rm $shortdir
	fi
	if [ "$dir" = "zyxel" ] ; then
	    ln -s /.etc/zyxel $shortdir
	else
	    ln -s /.etc $shortdir
	fi
    done
    
    mkdir -p ${chdir}/proc
    mount -t proc proc ${chdir}/proc
    mkdir -p ${chdir}/sys
    mount -t sysfs sys ${chdir}/sys

    mkdir -p ${chdir}/home
    mount --bind ${FFPROOT}/home ${chdir}/home

    CHROOTDIR=${chdir}
    return 0
}

# ========================================================
# function ClassicBindAll
# All firmware root directories are bind mounted on ffproot
# ========================================================
ClassicBindAll()
{
    # mount directories for chroot
    tomount="dev var tmp bin sbin lib usr"

    for arg in bind rbind
    do
	for tm in ${tomount}
	do
	    mkdir -p ${tm}
	    mount --${arg} /${tm} ${tm}
	    if [ "$?" != "0" ] ; then
		return 1
	    fi
	done
	
	tomount=""

	if [ "$arg" = "bind" ] ; then
	    mkdir -p e-data
	    mkdir -p i-data
	    if SymlinkDir e-data ${chdir}
	    then
		if ! SymlinkDir i-data ${chdir}
		then
		    tomount="$tomount i-data"
		fi
	    else
		tomount="$tomount e-data i-data"
	    fi
	fi
    done

    echo Create convenience links /ffp/etc/original and /ffp/etc/zyxel
    for dir in original zyxel
    do
	local shortdir=ffp/etc/$dir
	if [ ! -d $shortdir ] ; then
	    rm $shortdir
	    mkdir -p $shortdir
	fi
	if [ "$dir" = "zyxel" ] ; then
	    mount --bind /etc/zyxel $shortdir
	else
	    mount --bind /etc $shortdir
	fi
    done

    mkdir -p proc
    mkdir -p sys
    mount -t proc none proc
    mount -t sysfs none sys
    CHROOTDIR=${FFPROOT}
    return 0
}

# ================================================
# function FinalBindAll
# 'poor man's chroot'
# FFP is chrooted with hardly any connection to the firmware
# ================================================
FinalBindAll()
{
    CopyTree cp /dev ${FFPROOT}/dev
    if ! [ -d ${FFPROOT}/dev/pts ] ; then
        mkdir -p ${FFPROOT}/dev/pts
    fi
    mount -t devpts devpts ${FFPROOT}/dev/pts
    for dir in i-data e-data
    do
	mkdir -p ${FFPROOT}/${dir}
	CopyTree cp /${dir} ${FFPROOT}/${dir} 
    done
    mount -t proc proc proc
    mount -t sysfs sys sys
    CHROOTDIR=${FFPROOT}
}

WriteParams()
{
# ===================================
# Added 2012-08-26 
# Sometimes $FIRMWARE has quotes around, sometimes it doesn't.
# So add them, and filter for double quotes.
    if [ -x /bin/tr ] ; then
	echo NASTYPE=\"${NASTYPE}\" | tr -s '\042\042' '\042' >$1
	echo FIRMWARE=\"${FIRMWARE}\" | tr -s '\042\042' '\042' >>$1
    else
# ...And if 'tr' is not available, lump it
	echo NASTYPE=\"${NASTYPE}\"  >$1
	echo FIRMWARE=\"${FIRMWARE}\" >>$1
    fi
    echo FFPROOT=${FFPROOT} >>$1
    echo CHROOTDIR=${CHROOTDIR} >>$1
    echo FFPTYPE=${FFPTYPE} >>$1
}

# ===========================================
# function IsFFPinstalled()
# Checks for existance of FFP, to decide to 
# download or not
# ===========================================
IsFFPinstalled()
{
    if [ -r ffp/bin/busybox ] ; then
	return 0
    fi
    if [ -r ffp/bin/busybox-netutils ] ; then
	return 0
    fi
    return 1
}

# ===========================================
# function StartFFP()
# main function. Will download ffp if necessary
# create an environment for chrooting,
# chroot ffp
# ===========================================
StartFFP()
{
    # create rootdirectory
    if [ -r rootfs.tgz ]; then
	echo -ne Extract rootfs.tgz...
        tar -xzf rootfs.tgz
	rm rootfs.tgz
	echo Done
    fi

    CheckForFfpTarball
    
    if ! IsFFPinstalled ; then
	TryToDownload

        if ! IsFFPinstalled ; then
	    echo Could not download fun_plug.tgz for some reason
	    if [ "${FFPTYPE}" = "zypkg" ] || [ "${FFPTYPE}" = "lifeline" ] ; then
		echo FAIL!
		exit 1
	    fi
	    echo Will start polling for the existance forever in the share containing ffproot...
	    while ! IsFFPinstalled ; do
		sleep 10
		CheckForFfpTarball
	    done
	fi
	
	if [ ! -f ${FFPROOT}/ffp/etc/ffp.version ] ; then
	    if [ -f ${FFPROOT}/ffp/etc/ffp-version ] ; then
		. ${FFPROOT}/ffp/etc/ffp-version
		FFPVERSION=${FFP_VERSION}/${FFP_ARCH}
	    else
		FFPVERSION=EDIT_ME
	    fi
	    echo FFPVERSION=${FFPVERSION} > ${FFPROOT}/ffp/etc/ffp.version
	fi
	
	if IsFFPinstalled ; then
	    if [ "${FFPENABLE_SSH:-auto}" = "auto" ] ; then 
		if [ "${FFPTYPE}" = "zypkg" ] ; then
		    FFPENABLE_SSH=yes
		elif [ "${FFPTYPE}" = "lifeline" ] ; then
		    FFPENABLE_SSH=yes
		else
		    FFPENABLE_SSH=no
		fi
	    fi
	
	    if [ "${FFPENABLE_SSH}" = "yes" ] ; then
		chmod a+x ${FFPROOT}/ffp/start/sshd.sh
	    fi
	fi 
    fi

    echo -ne Try to create a /ffp symlink...
    if [ -h /ffp ] ; then
	rm /ffp
    fi
    
    if ln -s ${FFPROOT}/ffp /ffp
    then
	echo done
    else
        FFPCHROOT=yes
	echo "Failed. Read only rootfs?"
    fi
   
    if [ "${FFPCHROOT:-auto}" = "auto" ] ; then
	echo -ne "To chroot or not to chroot..."
	if [ -f ${FFPROOT}/ffp/etc/ffp.version ] ; then
	    . ${FFPROOT}/ffp/etc/ffp.version
	elif [ -f ${FFPROOT}/ffp/etc/ffp-version ] ; then
	    . ${FFPROOT}/ffp/etc/ffp-version
	    FFPVERSION=${FFP_VERSION}/${FFP_ARCH}
	fi
	
	
	case ${FFPVERSION} 
	in
	    0.7.0/arm)
		echo "FFP 0.7.0/arm, no"
		FFPCHROOT=no
		;;
	    0.7/arm)
		echo "FFP 0.7/arm, no"
		FFPCHROOT=no
		;;
	    0.6)
		echo "FFP 0.6, no"
		FFPCHROOT=no
		;;
	    *)
		echo ${FFPVERSION}, yes
		FFPCHROOT=yes
		;;
	esac
    fi

    if [ "${FFPCHROOT}" = "yes" ] ; then
        ClassicBindAll
	if [ "$?" != "0" ]; then
    	    echo "Bummer! Cannot bind directories. Will try the NSA320 hack to get it going"
	    NSA320BindAll
    	    if [ "$?" != "0" ]; then
		echo "The NSA320 hack doesn't work either. Will do the best I can to get it working"
		echo Possibilities are limited.
		FinalBindAll
	    fi
	fi

        echo Will chroot ffp on ${CHROOTDIR}
    else
	CHROOTDIR=""
    fi
    
    if [ "${FFPCHROOT}" = "no" ] && [ "${FFPCHANGEROOTSHELL:-yes}" = "yes" ] ; then
	    echo Change root shell
	    sed -i '/^root:/s/:\/bin\/sh/:\/ffp\/bin\/sh/g' /etc/passwd
	    echo Change admin shell
	    sed -i '/^admin:/s/:\/bin\/sh/:\/ffp\/bin\/sh/g' /etc/passwd
    fi
    
    # Generate helper file(s)
    local basedir=${CHROOTDIR}
    for hfnr in 1 2
    do
	local hd=${basedir}${HELPERDIR}
	local hf=${hd}${HELPERFILE}
	echo -ne Generate helper file ${hfnr} ${hf}... 
	mkdir -p ${hd}
	chmod 755 ${hd}
	WriteParams ${hf}
	echo done
	basedir=""
    done

    # Extend /ffp/etc/profile
    if ! cat ${CHROOTDIR}/ffp/etc/profile | grep FFPVERSION >/dev/null
    then
cat >>${CHROOTDIR}/ffp/etc/profile <<_EOF_
. ${HELPERDIR}${HELPERFILE}
. /ffp/etc/ffp.version
echo "== Welcome at FFP version \${FFPVERSION}, running on \${NASTYPE} \` hostname \`, firmware \${FIRMWARE} =="
_EOF_
    fi

    CHROOTCOMMAND=""
    if [ "" != "${CHROOTDIR}" ] ; then
	if [ "$NASTYPE" != "NSA310_TDC" ] ; then
	    # start helper
	    echo Start execute_outside_chroot Helper...
	    $0 InvokeHelper -c ${CHROOTDIR} &
	    sleep 1
	    echo done
	fi
	CHROOTCOMMAND="chroot ${CHROOTDIR}"
    fi

    # Copy /i-data/md0 symlink to /mnt/HD_a2 to make the ZyXEL more compatible with the DNS-323
    if [ -h /i-data/md0 ] ; then
	echo Generate convenience symlink /mnt/HD_a2
	mkdir -p ${CHROOTDIR}/mnt
	CopySymlink /i-data/md0 ${CHROOTDIR}/mnt/HD_a2
    fi

    # Generate /mnt/HD_a2 symlink to /mnt/pools/A/A0 to make the Iomega more compatible with the DNS-323
    if [ -d /mnt/pools/A/A0 ] ; then
	ln -s /mnt/pools/A/A0 /mnt/HD_a2 
    fi
    
    # Create an /etc/mtab symlink, to make df work
    if [ ! -f /etc/mtab ] ; then
	echo Create symlink /etc/mtab to /proc/mounts
	ln -s /proc/mounts /etc/mtab
    fi

    # run fun_plug local, if present
    if [ -x ${CHROOTDIR}/ffp/etc/fun_plug.local ]; then
	${CHROOTCOMMAND} /ffp/etc/fun_plug.local
    fi
 
    # start services
    if [ -x ${CHROOTDIR}/ffp/etc/rc ]; then
    	echo exec ${CHROOTCOMMAND} /ffp/etc/rc
        exec ${CHROOTCOMMAND} /ffp/etc/rc
    fi

    exit 0
}

# ============================================
# function StopFFP()
# terminates all ffp processes,
# unmounts all ffp mounts
# ============================================
StopFFP()
{
    if [ "${CHROOTDIR}" = "" ]; then
	if [ -x /ffp/etc/rc ]; then
	    echo /ffp/etc/rc stop
	    /ffp/etc/rc stop
	fi
	
	
	# ====================================
	# Added 20120825:
	# Change back root shell 
	# stopping FFP
	if [ "${FFPCHANGEROOTSHELL:-yes}" = "yes" ] ; then
	    echo Revert root shell
	    sed -i '/^root:/s/\/ffp\/bin\/sh/\/bin\/sh/g' /etc/passwd
	    echo Revert admin shell
	    sed -i '/^admin:/s/\/ffp\/bin\/sh/\/bin\/sh/g' /etc/passwd
	fi
    else
        cd /
	# stop services
	if [ -x ${CHROOTDIR}/ffp/etc/rc ]; then
	    echo /ffp/etc/rc stop
	    chroot ${CHROOTDIR} /ffp/etc/rc stop
	fi
    fi
    sleep 1

    message="Try to kill all remaining chrooted processes"
    for killtype in SIGTERM SIGKILL
    do    
	# List of all programs pid
	tokill=` ps | awk '{print $1}' `
	for pf in $tokill
	do
	    if [ -f /proc/$pf/maps ]; then
		if grep ${CHROOTDIR}/ffp/lib /proc/$pf/maps >/dev/null ; then                    
		    # Program has some chrooted libraries loaded
		    if [ "${message}" != "" ] ; then
			echo ${message} ; message=""
		    fi
            	    echo "      " ` cat /proc/$pf/cmdline `
            	    kill -${killtype} $pf
        	fi  
    	    fi          
	done                
        sleep 1
	message="Try a bit harder"
    done

    if [ "${CHROOTDIR}" = "" ] ; then
	exit 0
    fi

    echo Unmount all FFP mounts...
    local tounmount="none"
    while [ "${tounmount}" != "" ]; do
    	local lasttounmount=${tounmount}

	# While loop because mounts are nested.
	for unmount in ${tounmount} ; do
	    if [ "$unmount" != "none" ] && umount $unmount 2>/dev/null ; then
	        echo "      " $unmount
	    fi  
	done
	tounmount=` cat /proc/mounts | grep ${CHROOTDIR} | awk '{print $2}' 2>/dev/null `

	# filter the mounts which are symlinked to
	for filter in etc e-data i-data ; do
	    if [ -h /$filter ] ; then
		local target=` ls -l /$filter | awk -F\  '{print $NF}' `
		local filtered=""
		for unmount in ${tounmount} ; do
		    if ! echo $unmount | grep $target >/dev/null
		    then
			filtered="$filtered $unmount"
		    fi
		done
		tounmount=$filtered
	    fi
	done
	
	if [ "${tounmount}" = "${lasttounmount}" ] ; then
		tounmount=""
	fi
    done
    echo done

    rm -rf ${CHROOTDIR}/${HELPERDIR}
    rm -rf ${HELPERDIR}

    exit 0
}

# ==================================================
# function InvokeHelper()
# Helperfunction to be able to run programs outside the chroot.
# It will sleep until it's woken up by a SIGCONT,
# then in will execute scripts in HELPERDIR
# ==================================================
HelperLog()
{
    echo ` date +"%b %d %H:%m:%S" `: Helper script: $*
}

InvokeHelper()
{
    local hd=${CHROOTDIR}${HELPERDIR}
    local hf=${hd}${HELPERFILE}

    mv ${hf} ${hf}.tmp
    cat ${hf}.tmp | grep -v HELPERPID >${hf}
    rm ${hf}.tmp

    if [ "$1" = "stop" ]; then
	exit 0
    fi

    echo HELPERPID=$$ >>${hf}

    HelperLog started with pid $$
    
    while [ true ]; do
        trap "HelperLog Stop" SIGTERM
	if [ ! -d ${hd} ] ; then
	    trap "HelperLog Terminated" SIGTERM
	    exit 0
	fi
	
	local scripts=` ls ${hd} | grep helper | grep .sh `
	if [ "${scripts}" != "" ] ; then
	    trap "HelperLog Terminated" SIGTERM

	    for script in ${scripts}
	    do
		HelperLog Executing ${script}
		tmpfile=/tmp/$$` date +%s `
		mv ${hd}${script} ${tmpfile}
		chmod a+x ${tmpfile}
		${tmpfile}
	    	rm ${tmpfile}
	    done
	elif [ -f ${hf} ] ; then

	    HelperLog No more jobs to do\; Sleeping
	    kill -SIGSTOP $$
	else
		HelperLog FFP stopped\; Exiting
		exit 0
	
	fi
    done
}

PrintHelp()
{
    echo Usage: $0 command arguments
    echo "   command is one of:"
    echo "       StartFFP"
    echo "       StopFFP"
    echo "       InvokeHelper"
    echo "       Help"
    echo "   arguments:"
    echo "       -f firmware"
    echo "       -n nastype"
    echo "       -r ffproot"
    echo "       -c chrootdir"
    echo "       -t type (stick,zypkg,lifeline)"
}

ReadParams()
{
    local chrootspecified=0

    shift
    while [ "$1" != "" ] ; do
	case $1 in
	    -f)
		FIRMWARE=$2
		shift
		;;
	    -n)
		NASTYPE=$2
		shift
		;;
	    -h)
		PrintHelp
		shift
		;;
	    -r)
		FFPROOT=$2
		shift
		;;
	    -c)
		chrootspecified=1
		CHROOTDIR=$2
		shift
		;;
	    -t)
		FFPTYPE=$2
		shift
		;;
	    *)
		echo Unknown option $1
		;;
	esac
	shift
    done

    # =====================================
    # Added 20120825: for stopping FFP 
    # a better check of CHROOTDIR was
    # required
    if [ $chrootspecified -eq 0 ] && [ "" = "${CHROOTDIR}" ]; then
	CHROOTDIR=${FFPROOT}
    elif [ -z ${CHROOTDIR} ] ; then
	CHROOTDIR=""
    fi

    hf=${CHROOTDIR}${HELPERDIR}${HELPERFILE}
    if [ -r ${hf}  ]; then
        . ${hf}
    fi
}

# ================================================
# Entrypoint of the script
# ================================================
echo Starting $0 version ${SCRIPTVERSION} at ` date `...
SplitPath $0
FFPROOT=${SPLITTEDPATH}

if [ $# -eq 0 ]; then
    PrintHelp
    exit 0
fi

case $1 in
	StartFFP)
		ReadParams "$@"
		StartFFP
		;;
	StopFFP)
		ReadParams "$@"
		StopFFP
		;;
	InvokeHelper)
		ReadParams "$@"
		InvokeHelper
		;;
	Test)
		ReadParams "$@"
		tmpfile=/tmp/` date +%s%N `
		WriteParams ${tmpfile}
		cat ${tmpfile}
		rm ${tmpfile}
		;;
	*)
		if [ $# -eq 3 ]; then	
			# Assume this is an 'old style' call
			FFPROOT=$1
			NASTYPE=$2
			FIRMWARE=$3
			StartFFP
		else
			PrintHelp
		fi
		;;
esac

